#!/usr/bin/env python

from optparse import OptionParser
from van.pydeb import bin_to_py

import packages
import visualize


def build_target_kgs(old_eggs, ztk_eggs):
    target_kgs = packages.Eggs()

    for package in ztk_eggs:
        name = package.name
        version = package.simple_ver
        if package in old_eggs and not old_eggs[package].simple_ver:
            version = '' # keep unspecified versions unspecified
        target_kgs.add(packages.EggDesc(package=name, version=version))

    for package in old_eggs:
        if package not in target_kgs:
            target_kgs.add(package)

    return target_kgs


def write_schooltool_kgs(kgs, filename_out, overrides={}, intro='', sort_key=None):
    file_out = open(filename_out, 'w')
    if intro.strip():
        file_out.write(intro.strip() + '\n\n')
    file_out.write('[versions]\n')

    modified_kgs = packages.Eggs()
    modified_kgs.update(kgs)
    for name, info in overrides.items():
        modified_kgs[name]['Version'] = info['Version']

    empty_versions = packages.Eggs()
    empty_versions.update([pkg for pkg in modified_kgs
                           if not pkg.simple_ver])

    if len(empty_versions):
        file_out.write('\n# We want the latest features for development\n')
        for package in sorted(empty_versions, key=sort_key):
            file_out.write('%s =\n' % package.name)
        file_out.write('\n')

    file_out.write('# Dependencies\n\n')
    for package in sorted(modified_kgs, key=sort_key):
        if package in empty_versions:
            continue
        name = package.name
        if name not in overrides:
            file_out.write('%s = %s\n' % (name, package.simple_ver))
        else:
            comments =  overrides[name].get('Comments', '')
            for line in comments.strip().splitlines():
                file_out.write('# %s\n' % line.strip())
            file_out.write('#%s = %s\n' % (name, kgs[name].simple_ver))
            file_out.write('%s = %s\n' % (name, overrides[name]['Version']))

    file_out.close()


def migrate_schooltool(st_kgs_name, ztk_kgs_name,
                       filename_out, intro_text='',
                       overrides={},
                       names=None,
                       verbose=True):

    st_eggs = packages.Eggs(st_kgs_name, names)
    ztk_eggs = packages.Eggs(ztk_kgs_name, names)

    target_kgs = build_target_kgs(st_eggs, ztk_eggs)

    sort_key = packages.namespace_sort_key([
        '', 'zope.', 'zope.app.', 'zc.', 'z3c.'])

    write_schooltool_kgs(
        target_kgs, filename_out, overrides=overrides,
        intro=intro_text, sort_key=sort_key)

    new_kgs = packages.Eggs(filename_out)

    if verbose:
        log = visualize.Legendator(ommit_prefixes=['SAME'])

        visualize.print_header('Migrating SchoolTool KGS', level=2)

        print 'Migrating from', repr(st_kgs_name), 'to', repr(ztk_kgs_name)
        print 'Resulting kgs stored in', repr(filename_out)

        visualize.print_header('Changes in schooltool KGS')
        visualize.print_diff(
            st_eggs, new_kgs, sort_key=sort_key,
            log=log, with_legend=True, by_prefix=True)

        visualize.print_header('Changes from ZTK')
        visualize.print_diff(
            ztk_eggs, new_kgs, sort_key=sort_key,
            log=log, with_legend=True, by_prefix=True)

    return new_kgs


st_14_kgs_intro = '''
# SchoolTool 1.4 KGS
# ==================
#
# Versions for http://launchpad.net/schooltool/development series
#
# Put the following in your buildout.cfg to use::
#
#   [buildout]
#   extends = http://ftp.schooltool.org/schooltool/1.4/versions.cfg
#   versions = versions
#   find-links = http://ftp.schooltool.org/schooltool/1.4
#
# Setup parameters in setup.py::
#
#   dependency_links=['http://ftp.schooltool.org/schooltool/1.4']
#
'''


st_14_overrides = {
    'zope.html': dict(
        Version='1.2.0',
        Comments="""
        Downgrade zope.html until confirmed that new CK editor works with ST
        or required porting is done.
        """),
    'lxml': dict(
        Version='2.1.5',
        Comments="""
        Older version from Karmic.
        """),
    'zope.app.exception': dict(
        Version='3.6.0',
        Comments="""
        Should have been released long ago
        """),
    'zope.catalog': dict(
        Version='3.8.1',
        Comments="""Already have this in Lucid"""),
    'zope.lifecycleevent': dict(
        Version='3.6.0',
        Comments="""Already have this in Lucid"""),
    'zope.app.principalannotation': dict(
        Version='3.7.0',
        Comments="""Updated dependencies, uploaded to Karmic"""),
    'zope.app.container': dict(
        Version='3.8.1',
        Comments="""Fixed tests with newer packages"""),
    'zope.app.locales': dict(
        Version='3.5.2',
        Comments="""Fixed tests, uploaded to Karmic"""),
    'zope.configuration': dict(
        Version='3.7.1',
        Comments="""Already have this in Lucid"""),
    'zope.container': dict(
        Version='3.10.1',
        Comments="""Already have this in Lucid"""),
    'zope.location': dict(
        Version='3.8.2',
        Comments="""Already have this in Lucid"""),
    'zope.mimetype': dict(
        Version='1.2.0'),
    'zope.minmax': dict(
        Version='1.1.2'),
    'zope.schema': dict(
        Version='3.6.1',
        Comments="""Already have this in Lucid"""),
    'zope.tales': dict(
        Version='3.5.0',
        Comments="""Already have this in Lucid"""),
    'zope.traversing': dict(
        Version='3.12.0',
        Comments="""Reduced test dependencies, already in Lucid"""),
    'zope.app.appsetup': dict(
        Version='3.13.0'),
    'zope.app.principalannotation': dict(
        Version='3.7.0'),
    'zope.app.localpermission': dict(
        Version='3.7.2'),
    'zc.resourcelibrary': dict(
        Version='1.3.1'),
    'zope.app.pagetemplate': dict(
        Version='3.10.1',
        Comments="""Already have this in Lucid"""),
    }

if __name__ == '__main__':

    parser = OptionParser()
    (options, args) = parser.parse_args()

    if len(args) >= 1:
        target_ztk = args[0]
    else:
        target_ztk = 'ztk-107027.cfg'

    names = open('package_list', 'r').read().splitlines()
    names_10 = open('package_list-1.0', 'r').read().splitlines()
    all_names = set()
    all_names.update(names)
    all_names.update(names_10)
    py_names = [bin_to_py(name) for name in all_names]

    migrate_schooltool('schooltool.kgs.cfg', target_ztk,
                       'versions.cfg', intro_text=st_14_kgs_intro,
                       overrides=st_14_overrides,
                       #names=py_names,
                       verbose=True)

