import apt_pkg # apt-get install python-apt
from pkg_resources import parse_version
from copy import copy
from van.pydeb import py_to_bin, bin_to_py

import visualize


APT_INITIALIZED = False
def require_apt():
    global APT_INITIALIZED
    if not APT_INITIALIZED:
        apt_pkg.InitSystem()
        APT_INITIALIZED = True

SHOW_GAUGES = True


class PackageDesc(dict):

    def __init__(self, package=None, version=None):
        super(PackageDesc, self).__init__()
        if package is not None:
            self['Package'] = package
        if version is not None:
            self['Version'] = version

    @property
    def name(self):
        return self['Package']

    @property
    def simple_ver(self):
        return self['Version']

    def cmp_ver(self, other):
        return cmp(parse_version(self.simple_ver),
                   parse_version(other.simple_ver))

    def __str__(self):
        return '%(Package)s (%(Version)s)' % self

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, str(self))

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name


class DebDesc(PackageDesc):

    def __init__(self, lines):
        entry = ''
        for line in lines:
            line = line.strip()
            if line[0].isalpha():
                i = line.find(':')
                entry = line[:i]
                line = line[i+1:].strip()
            if entry not in self:
                self[entry] = line
            else:
                self[entry] += '\n%s' % line

    @property
    def simple_ver(self):
        # crappy guess at simple package version
        version = self['Version'].split('-')[0]
        version = version.replace('~', '')
        if ':' in version:
            return version[version.find(':')+1:]
        return version

    def cmp_ver(self, other):
        if isinstance(other, DebDesc):
            require_apt()
            return apt_pkg.VersionCompare(self['Version'], other['Version'])
        else:
            return PackageDesc.cmp_ver(self, other)


class EggDesc(PackageDesc):
    pass


class Packages(set):

    def __copy__(self):
        cls = type(self)
        inst = cls()
        inst.update(self)
        return inst

    def __getitem__(self, item):
        if isinstance(item, str):
            name = item
        elif isinstance(item, PackageDesc):
            name = item.name
        else:
            raise NotImplemented
        for pkg in self:
            if pkg.name == name:
                return pkg
        raise KeyError(item)

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default

    def add(self, item):
        super(Packages, self).add(copy(item))

    def remove(self, item):
        item = self[item]
        super(Packages, self).remove(item)

    def update(self, other):
        for item in other:
            self.add(copy(item))


class Debs(Packages):

    def __init__(self, filename=None, names=()):
        if filename is not None:
            self.load(filename, names)

    def _read_block(self, stream):
        lines = []
        s = stream.readline().strip()
        while s:
            lines.append(s)
            s = stream.readline().strip()
        return lines

    def load(self, filename, names=()):
        """Load packages Debian apt style file"""
        self.clear()
        f = open(filename)

        global SHOW_GAUGES
        gauge = None
        if SHOW_GAUGES:
            f.seek(0, 2)
            f_len = f.tell()
            f.seek(0)
            if f_len > 1000000:
                gauge = visualize.Gauge(label=filename)
                gauge.reset(f_len)
        packages = Packages()
        lines = self._read_block(f)
        while lines:
            deb = DebDesc(lines)
            if not names or deb.name in names:
                self.add(deb)
            lines = self._read_block(f)
            if gauge:
                gauge.tick(f.tell())
        if gauge:
            gauge.done()
        f.close()


class Eggs(Packages):

    def __init__(self, buildout_cfg=None, names=()):
        if buildout_cfg is not None:
            self.load_versions(buildout_cfg, names)

    def _iter_buildout_cfg(self, stream):
        """Stupid buildout.cfg iterator.
        Yields line, package, version for every line in the stream.
        Yelds line, None, None when not inside [versions] entry point.
        """
        lines = stream.readlines()
        entrypoint = ''
        for s in lines:
            s = s.strip()
            if s.startswith('[') and s.endswith(']'):
                entrypoint = s[1:-1]
            if (entrypoint=='versions' and
                s and s[0].isalpha() and '=' in s):
                package, version = [e.strip() for e in s.split('=')]
            else:
                package = version = None
            yield (s, package, version)


    def load_versions(self, filename, names=()):
        """Load versions from [versions] entrypoint in buildout.cfg"""
        f = open(filename)
        for line, package, version in self._iter_buildout_cfg(f):
            if package is not None:
                egg = EggDesc(package=package, version=version)
                if not names or egg.name in names:
                    self.add(egg)
        f.close()


class MappedPackages(object):

    def __init__(self, packages):
        self.packages = copy(packages)
        self.update()

    def update(self, packages=[]):
        # WARNING: don't mix Eggs with Debs.
        if packages:
            self.packages.update(packages)
        self.by_name = dict([
            (self._make_name(package), package)
            for package in self.packages])
        if len(self.by_name) != len(self.packages):
            raise Exception('Some packages have identical names')

    def _make_name(self, item):
        name = None
        if isinstance(item, str):
            name = item
        elif isinstance(item, DebDesc):
            name = item.name
            if isinstance(self.packages, Eggs):
                name = bin_to_py(name)
        elif isinstance(item, EggDesc):
            name = item.name
            if isinstance(self.packages, Debs):
                name = py_to_bin(name)
        else:
            raise NotImplemented
        return name

    def __contains__(self, item):
        try:
            return self._make_name(item) in self.by_name
        except KeyError:
            return False

    def __getitem__(self, item):
        name = self._make_name(item)
        return self.by_name[name]

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default

    def __iter__(self):
        return iter(self.packages)


def namespace_sort_key(namespaces):
    """Build a sort key, by namespaces.
    e.g.: namespaces = ['', 'zope.', 'zope.app.', 'zc.', 'z3c.']
    """
    def key(pkg):
        best_score = len(namespaces) + 1
        most_specific = None
        name = pkg.name
        for score, namespace in enumerate(namespaces):
            if name.startswith(namespace):
                if (most_specific is None or
                    len(namespace) > len(most_specific)):
                    best_score = score
                    most_specific = namespace
        return (best_score, str(pkg))
    return key


def loose_namespace_sort_key(namespaces):
    """Build a sort key, by namespaces.
    e.g.: namespaces = ['', 'zope.', 'zope.app.', 'zc.', 'z3c.']
    """
    def key(pkg):
        best_score = len(namespaces) + 1
        most_specific = None
        name = pkg.name
        for score, namespace in enumerate(namespaces):
            if namespace in name:
                if (most_specific is None or
                    len(namespace) > len(most_specific)):
                    best_score = score
                    most_specific = namespace
        return (best_score, str(pkg))
    return key
