import sys
from copy import copy


CLUTTER_STDOUT = False


def print_to_std(text):
    global CLUTTER_STDOUT
    if not CLUTTER_STDOUT:
        return
    sys.stdout.write(text)
    sys.stdout.flush()


class Legendator(object):

    def __init__(self, legend=None, ommit_prefixes=None):
        self.legend = {}
        if legend is not None:
            self.legend.update(legend)
        self.ommit = ommit_prefixes
        self.items = []

    @property
    def prefix_len(self):
        if not self.legend:
            return 0
        return max([len(prefix) for prefix, desc in self.legend.values()])

    def clear(self):
        self.items = []

    def add(self, prefix, text, description=None):
        if self.ommit and prefix in self.ommit:
            return
        self.items.append((prefix, text))
        if prefix not in self.legend:
            self.legend[prefix] = (prefix, description)
        else:
            l_pref, l_desc = self.legend[prefix]
            if (not l_desc) and description:
                self.legend[prefix] = (l_pref, description)

    def print_legend(self):
        print 'Legend:'
        spacing = self.prefix_len
        for prefix, description in sorted(self.legend.values()):
            if prefix in self.ommit:
                continue
            lead_in = prefix + ' ' * (spacing - len(prefix)) + ' -'
            description = description or 'N/A'
            for line in description.strip().splitlines():
                print '  %s %s' % (lead_in, line.strip())
                lead_in = ' ' * len(lead_in)
        print ''

    def print_items(self, by_prefix=False):
        if by_prefix:
            exp = [(i[0], i[1], n) for n, i in enumerate(self.items)]
            items = [(i[0], i[1])
                     for i in sorted(exp, key=lambda i: (self.legend[i[0]][0], i[2]))]
        else:
            items = self.items
        spacing = self.prefix_len
        last_pfx = None
        for prefix, line in items:
            if by_prefix and last_pfx is not None and last_pfx != prefix:
                print ''
            disp_prefix, description = self.legend[prefix]
            print '%s%s %s' % (disp_prefix,
                               ' '*(spacing-len(disp_prefix)),
                               line.strip())
            last_pfx = prefix

    def print_totals(self):
        prefix_totals = dict([(prefix, 0) for prefix in self.legend])
        for prefix, line in self.items:
            prefix_totals[prefix] += 1
        print 'Total: %s' % ', '.join(
            ['%d %s' % (prefix_totals[prefix], self.legend[prefix][0])
             for prefix in prefix_totals
             if self.legend[prefix][0].strip()])


def print_header(header, packages=None, level=0):
    if level==0:
        wrappings = '\n%%s\n%s' % ('-'*len(header))
    if level==1:
        wrappings = '\n%%s\n%s' % ('='*len(header))
    if level==2:
        wrappings = '\n%s\n= %%s =\n%s' % ('='*(len(header)+4), '='*(len(header)+4))

    print wrappings % header

    if packages is not None:
        print len(packages), 'packages.\n'
    else:
        print ''


class Gauge(object):
    MAX_LEN = 40

    def __init__(self, maximum=0, label=''):
        self.label = label
        self.reset(maximum)

    def reset(self, maximum):
        self.maximum = max(1, maximum)
        self.step = 1
        while self.maximum / self.step > self.MAX_LEN:
            self.step = self.step + 1
        self.value = -1

    def tick(self, value):
        value = max(0, min(value, self.maximum))
        i = value / self.step
        if i == self.value:
            return
        self.value = i

        t = self.maximum / self.step
        i = max(0, min(self.value, t))
        print_to_std('\r%s[%s%s]' % (
            self.label and (self.label+' ') or '',
            '.'*i, ' '*(t-i)))

    def done(self):
        self.tick(self.maximum)
        print_to_std('\n')


def print_diff(first, second, log=None,
               with_legend=True, by_prefix=True, sort_key=None):
    if log is None:
        log = Legendator()
    log.clear()

    both = set(first)
    for package in second:
        if package not in first:
            both.add(package)

    both = sorted(both, key=sort_key)
    both_len = len(both)
    gauge = None
    if both_len > 1000:
        gauge = Gauge(both_len, 'Diffing (%s total):' % both_len)

    for n, package in enumerate(both):
        in_first = first.get(package)
        in_second = second.get(package)

        if gauge:
            gauge.tick(n)

        line = '%s -> %s' % (in_first or '', in_second or '')
        if not in_first:
            log.add('NEW', line, description="New package")
        elif not in_second:
            log.add('REM', line, description="Removed package.")
        else:
            res = in_first.cmp_ver(in_second)
            if res < 0:
                log.add('UP', line, description="Package was upgraded.")
            elif res > 0:
                log.add('DOWN', line, description="Package was downgraded.")
            else:
                log.add('SAME', line, description="Package version unchanged.")

    if gauge:
        gauge.done()

    if with_legend:
        log.print_legend()
    log.print_items(by_prefix=by_prefix)
    print ''
    log.print_totals()
    log.clear()
